-- InRelief: a graphics engine for volumetric pixel art.
-- 2018-07-16 No Time To Play <https://notimetoplay.org/>

local Camera = { x = 0, y = 0, z = 0, f = 1, px = 0, py = 0, scale = 0 }

function Camera:new(cam)
	cam = cam or {}
	setmetatable(cam, self)
	self.__index = self
	return cam
end

function Camera:project(x, y, z)
	x = x - self.x
	y = y - self.y
	z = z - self.z
	if z ~= 0 then
		self.scale = self.f / z
		self.px = x * self.scale
		self.py = y * self.scale
	else
		self.scale = 0
		self.px = 0
		self.py = 0
	end
end

local function zcompare(a, b) return a.z > b.z end

local Sprite = { _hotspot_x = 0, _hotspot_y = 0, _hotspot_z = 0 }

function Sprite:new(pixels, colormap)
	local sprite = {
		height = #pixels, -- Let it crash if zero.
		width = #pixels[1], -- Same as above.
		depth = 0, -- Used for culling.
		
		voxels = {} -- Store as sparse array.
	}
	
	for y = 1, sprite.height do
		for x = 1, sprite.width do
			local z = pixels[y][x]
			if z > 0 then -- Zero means transparent voxel.
				local v = {x = x - 1, y = y - 1, z = -z,
					color = colormap[z]}
				table.insert(sprite.voxels, v)
				if z > sprite.depth then
					sprite.depth = z
				end
			end
		end
	end
	table.sort(sprite.voxels, zcompare)

	setmetatable(sprite, self)
	self.__index = self
	return sprite
end

function Sprite:set_hotspot(x, y, z)
	z = z or 0
	for i, v in ipairs(self.voxels) do
		v.x = v.x + self._hotspot_x - x
		v.y = v.y + self._hotspot_y - y
		v.z = v.z + self._hotspot_z - z
	end
	self._hotspot_x = x
	self._hotspot_y = y
	self._hotspot_z = z
end

function Sprite:render_at(x, y, z, cam, render)
	for i = 1, #self.voxels do -- Faster than ipairs here.
		local v = self.voxels[i]
		cam:project(x + v.x, y + v.y, z + v.z)
		render(cam.px, cam.py, cam.scale, v.color)
	end
end

local SpriteInstance = {}

function SpriteInstance:new(sprite, x, y, z)
	local inst = {sprite = sprite, x = x, y = y, z = z}
	
	setmetatable(inst, self)
	self.__index = self
	return inst
end

function SpriteInstance:render_through(camera, fn)
	if self.z - self.sprite.depth > camera.z then
		self.sprite:render_at(
			self.x, self.y, self.z, camera, fn)
	end
end

local DisplayList = {}

function DisplayList:new()
	local dlist = {sprites = {}}
	setmetatable(dlist, self)
	self.__index = self
	return dlist
end

function DisplayList:insert(sprite, x, y, z)
	local s = SpriteInstance:new(sprite, x, y, z)
	table.insert(self.sprites, s)
	return s
end

function DisplayList:zsort()
	table.sort(self.sprites, zcompare)
end

function DisplayList:render_through(camera, fn)
	for i, s in ipairs(self.sprites) do
		s:render_through(camera, fn)
	end
end

return {
	zcompare = zcompare,
	Camera = Camera,
	Sprite = Sprite,
	SpriteInstance = SpriteInstance,
	DisplayList = DisplayList
}
